/* app.c
 *
 * Copyright (C) 1999 Havoc Pennington
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

#include "app.h"
#include "grid.h"
#include "menus.h"
#include "palette.h"
#include "edit.h"
#include <gtkpalette.h>
#include <gtkpastebuffer.h>
#include <iestatus.h>
#include <config.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <libart_lgpl/art_alphagamma.h>
#include <libart_lgpl/art_rgb_rgba_affine.h>
#include <libart_lgpl/art_filterlevel.h>
#include <libart_lgpl/art_affine.h>

/* For the icon setting routines */
#include <gdk/gdkprivate.h>
#include <gdk/gdkx.h>
#include <X11/Xlib.h>

/* The global paste buffer */
extern GtkObject *paste_buffer;

/* Keep a list of all open application windows */
static GSList* app_list = NULL;

static gint delete_event_cb(GtkWidget* w, GdkEventAny* e, gpointer data);
static void fg_color_set(GnomeColorPicker *cp, guint r, guint g, guint b, guint a, Grid* grid);
static void preview_expose(GtkWidget* da, GdkEventExpose* event, Grid* grid);
static void mini_expose(GtkWidget* da, GdkEventExpose* event, Grid* grid);
void item_added_cb (GtkPasteBuffer *pbuffer,
		    gpointer data);
void item_removed_cb (GtkPasteBuffer *pbuffer,
		      gpointer data);
void ie_set_mini_icon (GdkWindow *window, GdkPixbuf *pb);
void colour_changed (GtkPalette *palette,
		     gint selection,
		     Grid *grid);

void
item_added_cb (GtkPasteBuffer *pbuffer, 
	       gpointer data)
{
  GtkWidget *app = GTK_WIDGET (data);
  AppMenus *am;

  am = gtk_object_get_data (GTK_OBJECT (app), "menu-widgets");

  gtk_widget_set_sensitive (am->paste, TRUE);
  gtk_widget_set_sensitive (am->b_paste, TRUE);

  if (gtk_object_get_data (GTK_OBJECT (app), "selection"))
    gtk_widget_set_sensitive (am->paste_into, TRUE);
}

void
item_removed_cb (GtkPasteBuffer *pbuffer,
		 gpointer data)
{
  GtkWidget *app = GTK_WIDGET (data);
  AppMenus *am;

  if (GTK_BUFFER (pbuffer)->buffer_size != 0)
    return;

  am = gtk_object_get_data (GTK_OBJECT (app), "menu-widgets");
  gtk_widget_set_sensitive (am->paste, FALSE);
  gtk_widget_set_sensitive (am->b_paste, FALSE);
  gtk_widget_set_sensitive (am->paste_into, FALSE);
}

void
colour_changed (GtkPalette *palette,
		gint selection,
		Grid *grid)
{
	gushort rs, gs, bs, as;
	guint r, g, b;

	gtk_palette_get_colour (palette, selection, &r, &g, &b);
	gnome_color_picker_set_i8 (GNOME_COLOR_PICKER (grid->picker),
				   r, g, b, 255);

	/* Fake the signal emmission */
	gnome_color_picker_get_i16 (GNOME_COLOR_PICKER (grid->picker),
				    &rs, &gs, &bs, &as);
	gtk_signal_emit_by_name (GTK_OBJECT (grid->picker), "color-set",
				 rs, gs, bs, as);
}

void
alpha_value_changed_cb (GtkAdjustment *adj,
			GtkLabel *label)
{
  gchar *text;
  gdouble percentage;

  percentage = (adj->value / 255) * 100;
  text = g_strdup_printf (_("Level: %d%%"), (gint)percentage);

  gtk_label_set_text (label, text);
  g_free (text);
}
  
GtkWidget* 
ie_app_new(guint width, guint height)
{
  GtkWidget* app;
  GtkWidget* status;
  GtkWidget* canvas;
  GtkWidget* preview;
  GtkWidget* dockitem;
  GtkWidget* sw;
  GtkWidget* palette;
  GtkWidget* cp;
  GtkWidget* frame;
  GtkWidget* alpha_vbox;
  GtkWidget* alpha;
  GtkWidget* alpha_label;
  GtkWidget* colour_vbox;
  GtkWidget* mini = NULL;
  GtkWidget* floater = NULL;
  Grid* grid;
  GtkObject *adj;
  GtkTooltips *tt;

  app = gnome_app_new(PACKAGE, _("Icon Editor"));
  tt = gtk_tooltips_new ();

  gtk_window_set_policy(GTK_WINDOW(app), FALSE, TRUE, FALSE);
  gtk_window_set_default_size(GTK_WINDOW(app), 575, 475);
  gtk_window_set_wmclass(GTK_WINDOW(app), "iconedit", "IconEdit");

  gtk_widget_push_visual(gdk_rgb_get_visual());
  gtk_widget_push_colormap(gdk_rgb_get_cmap());
  canvas = gnome_canvas_new_aa ();
  preview = gtk_drawing_area_new ();
  if (width <= 64 && height <= 64)
    {
      mini = gtk_drawing_area_new ();
      gtk_drawing_area_size (GTK_DRAWING_AREA (mini), 60, 20);
    }
  gtk_widget_pop_visual();
  gtk_widget_pop_colormap();

  sw = gtk_scrolled_window_new(NULL, NULL);

  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
                                 GTK_POLICY_AUTOMATIC,
                                 GTK_POLICY_AUTOMATIC);

  gtk_container_add(GTK_CONTAINER(sw), canvas);
  
  gnome_app_set_contents(GNOME_APP(app), sw);

  /* Add to dock */
  if (width <= 64 && height <= 64)
    {
      frame = gtk_frame_new(_("Preview"));
      gtk_container_add(GTK_CONTAINER(frame), preview);
      dockitem = gnome_dock_item_new ("Preview", GNOME_DOCK_ITEM_BEH_NORMAL);
      gtk_container_add (GTK_CONTAINER (dockitem), frame);
      
      gnome_dock_add_item (GNOME_DOCK (GNOME_APP (app)->dock),
			   GNOME_DOCK_ITEM (dockitem),
			   GNOME_DOCK_RIGHT, 0, 0, 0, FALSE);
      gtk_widget_show_all (dockitem);
    }
  else
    {
      /* Toplevel window to show preview */
      GtkWidget *scroller;
      floater = gtk_window_new (GTK_WINDOW_TOPLEVEL);
      gtk_window_set_title (GTK_WINDOW (floater), _("GNOME-Iconedit: Preview"));
      gtk_container_set_border_width (GTK_CONTAINER (floater), 2);
      scroller = gtk_scrolled_window_new (NULL, NULL);
      gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (scroller), preview);
      gtk_container_add (GTK_CONTAINER (floater), scroller);
      /* Attach the preview to the app, for destruction at the end */
      gtk_object_set_data (GTK_OBJECT (app), "preview-window", floater);
    }

  if (width <= 64 && height <= 64)
    {
      frame = gtk_frame_new (_("Mini-Preview"));
      gtk_container_add (GTK_CONTAINER (frame), mini);
      dockitem = gnome_dock_item_new ("Mini-Preview", GNOME_DOCK_ITEM_BEH_NORMAL);
      gtk_container_add (GTK_CONTAINER (dockitem), frame);
      gnome_dock_add_item (GNOME_DOCK (GNOME_APP (app)->dock),
			   GNOME_DOCK_ITEM (dockitem),
			   GNOME_DOCK_RIGHT, 0, 1, 0, FALSE);
      gtk_widget_show_all (dockitem);
    }

  status = ie_status_new (3, FALSE);
  gtk_widget_set_usize (IE_STATUS (status)->status->pdata[2], 50, -1);
  /* Naughty */
  gtk_box_set_child_packing (GTK_BOX (status), IE_STATUS (status)->status->pdata[2],
			     FALSE, FALSE, 0, GTK_PACK_START);

  gnome_app_set_statusbar(GNOME_APP(app), status);

  /* Maybe use a GtkVWrapBox once it's in the main GTK+ tree *shrug* 
     Might need GtkPalette to flip to be a horizontal widget as well */
  colour_vbox = gtk_vbox_new (FALSE, 2);
  gtk_container_set_border_width (GTK_CONTAINER (colour_vbox), 2);
  palette = gtk_palette_new (256);
  gtk_object_set_data (GTK_OBJECT (app), "palette", palette);

  /* Make this load a default palette - the generate default is a bit sucky */
  gtk_palette_set_palette_default (GTK_PALETTE (palette));
  
  gtk_box_pack_start (GTK_BOX (colour_vbox), palette, FALSE, FALSE, 1);

  cp = gnome_color_picker_new ();
  gtk_box_pack_start (GTK_BOX (colour_vbox), cp, FALSE, FALSE, 1);

  frame = gtk_frame_new (_("Alpha level"));

  alpha_vbox = gtk_vbox_new (FALSE, 0);
  alpha_label = gtk_label_new (_("Level:100%"));
  gtk_misc_set_alignment (GTK_MISC (alpha_label), 0, 0.5);

  gtk_box_pack_start (GTK_BOX (alpha_vbox), alpha_label, TRUE, TRUE, 0);
		      
  adj = gtk_adjustment_new (255.0, 
			    0.0,
			    255.0,
			    1.0,
			    10.0,
			    0.0);
  gtk_signal_connect (GTK_OBJECT (adj), "value-changed",
		      GTK_SIGNAL_FUNC (alpha_value_changed_cb), alpha_label);

  alpha = gtk_hscale_new (GTK_ADJUSTMENT (adj));
  gtk_scale_set_draw_value (GTK_SCALE (alpha), FALSE);
  
  gtk_box_pack_start (GTK_BOX (alpha_vbox), alpha, TRUE, TRUE, 0);

  gtk_container_add (GTK_CONTAINER (frame), alpha_vbox);
  gtk_box_pack_start (GTK_BOX (colour_vbox), frame, FALSE, FALSE, 1);

  /* Add to dock */
  dockitem = gnome_dock_item_new ("Colour-Palette", GNOME_DOCK_ITEM_BEH_NEVER_HORIZONTAL);
  gtk_container_add (GTK_CONTAINER (dockitem), colour_vbox);
  gnome_dock_add_item (GNOME_DOCK (GNOME_APP (app)->dock),
		       GNOME_DOCK_ITEM (dockitem),
		       GNOME_DOCK_RIGHT, 0, 2, 0, FALSE);
  gtk_widget_show_all (dockitem);

  grid = grid_new(app, canvas, preview, status, cp, 
		  GTK_ADJUSTMENT (adj), mini, width, height);

  gtk_object_set_data_full(GTK_OBJECT(app), "grid",
                           grid, (GtkDestroyNotify)grid_destroy);

  ie_install_menus_and_toolbar(app);

  /* Connect to the paste buffer's signals */
  gtk_signal_connect (paste_buffer, "item-added",
		      GTK_SIGNAL_FUNC (item_added_cb), 
		      app);

  gtk_signal_connect (paste_buffer, "item-removed",
		      GTK_SIGNAL_FUNC (item_removed_cb), 
		      app);

  gtk_signal_connect(GTK_OBJECT(app),
                     "delete_event",
                     GTK_SIGNAL_FUNC(delete_event_cb),
                     NULL);


  gtk_signal_connect(GTK_OBJECT(cp), "color_set",
                     GTK_SIGNAL_FUNC(fg_color_set),
                     grid);

  gtk_signal_connect(GTK_OBJECT(preview), "expose_event",
		     GTK_SIGNAL_FUNC(preview_expose),
		     grid);
  
  if (mini)
    gtk_signal_connect (GTK_OBJECT (mini), "expose_event",
			GTK_SIGNAL_FUNC (mini_expose),
			grid);

  gtk_signal_connect (GTK_OBJECT (palette), "selection-changed",
		      GTK_SIGNAL_FUNC (colour_changed), 
		      grid);

  app_list = g_slist_prepend(app_list, app);

  gtk_widget_show_all(sw);
  gtk_widget_show(status);
  if (floater)
    gtk_widget_show_all (floater);

  return app;
}

void
ie_update_grid (GtkWidget *app)
{
  Grid *g;

  g = gtk_object_get_data (GTK_OBJECT (app), "grid");

  grid_update_item (g);
}

void       
ie_app_close(GtkWidget* app)
{
  Grid *grid;
  GtkWidget *floater = NULL;
  gchar *filename;

  g_return_if_fail(GNOME_IS_APP(app));

  grid = gtk_object_get_data (GTK_OBJECT (app), "grid");
  floater = gtk_object_get_data (GTK_OBJECT (app), "preview-window");

  if (grid->dirty)
    {
      GtkWidget *dialog;
      gchar *message;
      
      filename = gtk_object_get_data (GTK_OBJECT (app), "filename");
      message = g_strdup_printf (_("%s has been changed.\nDo you want to save it?"),
				 filename ? filename: _("This icon"));
      dialog = gnome_message_box_new (message,
				      GNOME_MESSAGE_BOX_QUESTION,
				      GNOME_STOCK_BUTTON_YES,
				      GNOME_STOCK_BUTTON_NO,
				      GNOME_STOCK_BUTTON_CANCEL,
				      NULL);
      g_free (message);
      gtk_window_set_title (GTK_WINDOW (dialog), _("File not saved"));
      gnome_dialog_set_parent (GNOME_DIALOG (dialog), GTK_WINDOW (app));

      switch (gnome_dialog_run_and_close (GNOME_DIALOG (dialog)))
	{
	case 0:
	  save_cb (NULL, app);
	  break;
	case 1:
	  break;
	default:
	  return;
	}
    }

  /* Free the selection of the window */
  ie_selection_destroy (app);

  /* Disconnect from the paste baster. */
  gtk_signal_disconnect_by_data (paste_buffer, app);

  gdk_pixbuf_unref (grid->pixbuf);
  app_list = g_slist_remove(app_list, app);

  gtk_widget_destroy (app);
  if (floater)
    gtk_widget_destroy (floater);
  
  if (app_list == NULL)
    {
      /* No windows remaining */
      gtk_main_quit();
    }
}

static const char *tip[IET_END] = 
{
  N_("Select region mode."),
  N_("Paint brush mode."),
  N_("Flood fill mode."),
  N_("Eyedropper mode."),
  N_("Brush paster mode.")
};

void       
ie_app_set_tool(GtkWidget* app, IETool tool)
{
  Grid* g;

  g = gtk_object_get_data(GTK_OBJECT(app), "grid");
  
  g_assert(g != NULL);

  g->tool = tool;

  ie_status_set_status(IE_STATUS(g->status), _(tip[tool]), 1);
}

void
ie_app_set_mode (GtkWidget *app, IEMode mode)
{
  Grid *g;

  g = gtk_object_get_data (GTK_OBJECT (app), "grid");
  g_assert (g != NULL);

  g->mode = mode;
}

void       
ie_app_set_from_pixbuf(GtkWidget* app, GdkPixbuf *pb)
{
  Grid* g;

  g = gtk_object_get_data(GTK_OBJECT(app), "grid");

  g_assert(g != NULL);

  grid_set_from_pixbuf(g, pb);
}

GdkPixbuf* 
ie_app_to_pixbuf(GtkWidget* app)
{
  Grid* g;

  g = gtk_object_get_data(GTK_OBJECT(app), "grid");

  g_assert(g != NULL);

  return grid_to_pixbuf(g);
}

static gint 
delete_event_cb(GtkWidget* window, GdkEventAny* e, gpointer data)
{
  ie_app_close(window);

  /* Prevent the window's destruction, since we destroyed it 
   * ourselves with hello_app_close()
   */
  return TRUE;
}

static void 
fg_color_set(GnomeColorPicker *cp, guint r, guint g, guint b, guint a, Grid* grid)
{
  grid->current_color.red = r;
  grid->current_color.green = g;
  grid->current_color.blue = b;
  grid->current_color.pixel = 
    gdk_rgb_xpixel_from_rgb(((r/256) << 16) +
                            ((g/256) << 8) +
                            ((b/256)));
}

static void
mini_expose (GtkWidget *da,
	     GdkEventExpose *event,
	     Grid *grid)
{
  int i, x, y, bg[3][3], rowstride;
  GdkPixbuf *minipb;
  guchar *dest, *src, *pixels;
  guchar *buffer = g_new (guchar, 60*20*3);
  GdkGC *gc;

  if (grid->width >= 20 && grid->height >= 20)
    minipb = gdk_pixbuf_scale_simple (grid->pixbuf, 20, 20, ART_FILTER_NEAREST);
  else
    {
      minipb = grid->pixbuf;
      gdk_pixbuf_ref (minipb);
    }

  
  /* Setup the background colours */
  bg[0][0] = 0x00;
  bg[0][1] = 0x00;
  bg[0][2] = 0x00;
  bg[1][0] = 0xFF;
  bg[1][1] = 0xFF;
  bg[1][2] = 0xFF;
  bg[2][0] = da->style->bg[GTK_STATE_NORMAL].red / 256;
  bg[2][1] = da->style->bg[GTK_STATE_NORMAL].green / 256;
  bg[2][2] = da->style->bg[GTK_STATE_NORMAL].blue / 256;

  pixels = gdk_pixbuf_get_pixels (minipb);
  rowstride = gdk_pixbuf_get_rowstride (minipb);
  y = 0; 
  while (y < 20)
    {
      src = pixels + (y * rowstride);
      dest = buffer + (y * 180);;
      x = 0;
      while (x < 20)
	{
	  guchar r = *(src++);
	  guchar g = *(src++);
	  guchar b = *(src++);
	  guchar a = *(src++);
	  int v;
	  
	  for (i = 0; i < 3; i++)
	    {
	      v = bg[i][0];
	      dest[0 + (i * 60)] = v + (((r - v) * a + 0x80) >> 8);
	      v = bg[i][1];
	      dest[1 + (i * 60)] =  v + (((g - v) * a + 0x80) >> 8);
	      v = bg[i][2];
	      dest[2 + (i * 60)] =  v + (((b - v) * a + 0x80) >> 8);
	    }
	  dest += 3;
	  ++x;
	}
      ++y;
    }
    
  gc = gdk_gc_new (da->window);
  gdk_draw_rgb_image (da->window,
		      gc,
		      0, 0,
		      60, 20,
		      GDK_RGB_DITHER_NORMAL,
		      buffer,
		      60*3);
  g_free (buffer);
  gdk_gc_unref (gc);
  gdk_pixbuf_unref (minipb);
}

static void 
preview_expose(GtkWidget* da, GdkEventExpose* event, Grid* grid)
{
  gint w = da->allocation.width;
  gint h = da->allocation.height;
  gint x = 0, y = 0;
  gint extra_w = w - grid->width;
  gint extra_h = h - grid->height;
  guint bufsize = grid->width * grid->height * 3; 
  guchar* buf = g_malloc(bufsize);
  guchar *pixels;
  guchar *src;
  guchar *dest;
  guint i;

  pixels = gdk_pixbuf_get_pixels (grid->pixbuf);
  i = 0;
  src = pixels;
  dest = buf;
  while (i < bufsize/3)
    {
      guchar r = *src++;
      guchar g = *src++;
      guchar b = *src++;
      guchar a = *src++;
      int v;
	
      v = da->style->bg[GTK_STATE_NORMAL].red / 256;
      *(dest++) = v + (((r - v) * a + 0x80) >> 8);
      v = da->style->bg[GTK_STATE_NORMAL].green / 256;
      *(dest++) = v + (((g - v) * a + 0x80) >> 8);
      v = da->style->bg[GTK_STATE_NORMAL].blue / 256;
      *(dest++) = v + (((b - v) * a + 0x80) >> 8);

      ++i;
    }
  
  if (extra_w > 0)
    x = extra_w/2;
  
  if (extra_h > 0)
    y = extra_h/2;

  gdk_gc_set_clip_rectangle(da->style->black_gc,
			    &event->area);
  
  gdk_draw_rgb_image (da->window,
		      da->style->black_gc,
		      x, y,
		      grid->width, 
		      grid->height,
		      GDK_RGB_DITHER_NORMAL,
		      buf,
		      grid->width*3);
  
  gdk_gc_set_clip_rectangle(da->style->black_gc,
			    NULL);

  g_free (buf);
}

void
ie_app_dirty (GtkWidget *app)
{
  Grid *g;

  g = gtk_object_get_data (GTK_OBJECT (app), "grid");
  g->dirty = TRUE;

  /* Set the tasklist mini icon */
  ie_set_mini_icon (app->window, g->pixbuf);

  /* Update the preview */
  gtk_widget_queue_draw (g->preview);
  if (g->mini)
    gtk_widget_queue_draw (g->mini);
}

void
ie_app_clean (GtkWidget *app)
{
  Grid *g;

  g = gtk_object_get_data (GTK_OBJECT (app), "grid");
  g->dirty = FALSE;

  /* Set the tasklist mini icon */
  ie_set_mini_icon (app->window, g->pixbuf);

  /* Update the preview */
  gtk_widget_queue_draw (g->preview);
  if (g->mini)
    gtk_widget_queue_draw (g->mini);
}

void
ie_set_mini_icon (GdkWindow *window,
		  GdkPixbuf *pb)
{
  GdkPixmap *pmap, *bmap;
  GdkAtom icon_atom;
  GdkVisual *visual;
  long data[2];
  gint width, height;
  GdkGC *gc;

  /* Based on code from Nautilus by Anders Carlsson. */
  g_return_if_fail (window != NULL);
  g_return_if_fail (pb != NULL);

  visual = gdk_rgb_get_visual ();
  width = gdk_pixbuf_get_width (pb);
  height = gdk_pixbuf_get_height (pb);

  if (width > 20 || height > 20)
    return;

  pmap = gdk_pixmap_new (NULL, width, height, visual->depth);
  bmap = gdk_pixmap_new (NULL, width, height, 1);
  
  gc = gdk_gc_new (pmap);
  gdk_pixbuf_render_to_drawable (pb,
				 pmap, gc,
				 0, 0,
				 0, 0,
				 width, height,
				 GDK_RGB_DITHER_MAX,
				 0, 0);
  gdk_pixbuf_render_threshold_alpha (pb, bmap,
				     0, 0,
				     0, 0,
				     width, height,
				     127);
  gdk_gc_unref (gc);

  data[0] = ((GdkPixmapPrivate*) pmap)->xwindow;
  if (bmap)
    data[1] = ((GdkPixmapPrivate*) bmap)->xwindow;
  else
    data[1] = 0;

  icon_atom = gdk_atom_intern ("KWM_WIN_ICON", FALSE);
  gdk_property_change (window, icon_atom, icon_atom,
		       32, GDK_PROP_MODE_REPLACE,
		       (guchar*) data, 2);
}

void
ie_update_statusbar (Grid *grid, 
		     gint x,
		     gint y)
{
  gchar *message;
  guchar *pixels;
  guchar r, g, b, a;

  pixels = gdk_pixbuf_get_pixels (grid->pixbuf);

  pixels += (y * grid->rowstride) + (x * 4);
  r = pixels[0];
  g = pixels[1];
  b = pixels[2];
  a = pixels[3];

  message = g_strdup_printf (_("(%d,%d) - R:%d G:%d B:%d A:%d"),
			     x, y, r, g, b, a);
  ie_status_set_status (IE_STATUS (grid->status),
			message, 0);
  g_free (message);
}
